from copy import copy
from itertools import product, chain
from logging import info
from random import Random

from .board import Board
from .entity import Entity
from .errors import UndefinedSelectorError, UdebsExecutionError
from .interpret import Script, UdebsStr, _getEnv
from .utilities import no_recurse


# ---------------------------------------------------
#                 Main Class                        -
# ---------------------------------------------------

class Instance(dict):
    """
    Main instance class that represents the state space of the entire game.

    Note on naming convention:

    All methods beginning with 'get' return information about the game's state
    space and are not allowed to change the state itself.

    All methods beginning with 'control' or 'cast' manage some aspect of the state space
    and return true if changes were (probably) made and false otherwise.

    Other functions are convenience wrappers for public use.
    """

    def __init__(self, is_copy=False):
        if is_copy:
            return

        # definitions
        self.lists = {'group', 'effect', 'require'}
        self.stats = set()
        self.strings = set()
        # rlist and rmap are flags that indicate objects entities should inherit from.
        self.rlist = ['group']
        self.rmap = []

        # config
        self.name = 'Unknown'  # only effects what is printed when initialized
        self.logging = True  # Turns logging on and off
        self.revert = 0  # Determines how many steps should be saved in revert
        self.version = 1  # What version of Udebs syntax is used.
        self.seed = None  # Random seed for processing.
        self.immutable = False  # Determines default setting for entities immutability.
        self.auto_entity = True

        # time
        self.time = 0  # In game counter
        self.increment = 1  # how much time passes between game_loop iterations. Useful for pausing.
        self.cont = True  # Flag to determine if game_loop should continue
        self.next = None  # The next state a game_loop will return. Useful for resets and reverts in game_loop
        self.value = None  # Value of the game.

        # internal
        self.map = {}
        self.delay = []
        self.rand = Random()

        # Do not copy
        self.state = None

        super().__init__()

    def __bool__(self):
        return self.cont

    @no_recurse
    def __eq__(self, other):
        if not isinstance(other, Instance):
            return False

        for k, v in self.__dict__.items():
            if k not in ("state", "rand") and v != getattr(other, k):
                return False

        for k, v in self.items():
            if other[k] != v:
                return False

        if len(self) != len(other):
            return False

        return True

    def __ne__(self, other):
        return not self == other

    def __copy__(self):
        return self.copy()

    def copy(self, new=None) -> "Instance":
        if new is None:
            new = type(self)(True)

        for k, v in self.__dict__.items():
            if k not in {"delay", "map", "state"}:
                setattr(new, k, v)

        # Handle entities
        for name, entity in self.items():
            if entity.immutable:
                new[name] = entity
            else:
                new[name] = entity.copy()

        # Handle maps
        new.map = {}
        for name, map_ in self.map.items():
            new.map[name] = map_.copy()

        # Handle delays
        new.delay = []
        for delay in self.delay:
            env = copy(delay["env"])
            env["self"] = new
            env["storage"] = {k: new[v.name] for k, v in delay["env"]["storage"].items()}

            new.delay.append({
                "env": env,
                "ticks": delay["ticks"],
                "script": delay["script"],
            })

        new.state = None

        return new

    # ---------------------------------------------------
    #               Selector Function                  -
    # ---------------------------------------------------

    def getEntity(self, target):
        """
        Fetches the udebs entity object given a selector.

        Udebs selectors can take all of the following forms.

        * string - The name of the object we are looking for.
        * tuple - A tuple of the form (x, y, name) representing a location on a map.
        * list - A list containing other selectors.
        * None - Returns the default empty selector.

        .. code-block:: xml

            <i># identifier</i>

        """
        if isinstance(target, Entity):
            return target
        elif target is False or target is None or target == "bump":
            return self['empty']
        elif isinstance(target, str) and target in self:
            return self[target]
        elif isinstance(target, tuple):
            return self._getEntityTuple(target)
        elif isinstance(target, list):
            return [self.getEntity(i) for i in target]
        elif isinstance(target, UdebsStr):
            return self._getEntityUdebsStr(target)

        raise UndefinedSelectorError(target, "entity")

    def _getEntityUdebsStr(self, target):
        # Process new nameless.
        scripts = []
        if target[0] == "(":
            while target[0] == "(":
                target = target[1:-1]

            bracket = 0
            buf = []
            for char in target:
                if char == "(":
                    bracket += 1

                if char == ")":
                    bracket -= 1

                if not bracket and char == ",":
                    scripts.append(Script(UdebsStr("".join(buf)), self.version))
                    buf = []
                    continue

                buf.append(char)

            raw = UdebsStr("".join(buf))
            scripts.append(Script(raw, self.version))

        else:
            scripts.append(Script(target, self.version))

        self[target] = Entity(self, require=scripts, name=target, immutable=True)
        return self[target]

    def _getEntityTuple(self, target):
        if len(target) < 3:
            map_name = "map"
        else:
            map_name = target[2]

        try:
            name = self.map[map_name][target]
        except (IndexError, TypeError, KeyError):
            raise UndefinedSelectorError(target, "entity")

        unit = self[name]

        if not unit.loc:
            if len(target) < 3:
                target = (*target, map_name)
            unit = unit.copy(loc=target)

        return unit

    def getLocObject(self, target):
        if isinstance(target, tuple):
            if len(target) < 3:
                target = (*target, "map")
            return target
        return self.getEntity(target).loc

    def getMap(self, target="map"):
        """
        Fetches the map object caster currently resides on.

        .. code-block:: xml

            <i>caster [$caster] MAP</i>
        """
        if isinstance(target, tuple):
            target = target[2] if len(target) == 3 else "map"
        if not target:
            return None
        elif isinstance(target, str) and target in self.map:
            return self.map[target]
        elif isinstance(target, Board):
            return target
        elif isinstance(target, Entity):
            return self.map[target.loc[2]]
        else:
            raise UndefinedSelectorError(target, "map")

    # ---------------------------------------------------
    #                 Time Management                  -
    # ---------------------------------------------------

    def _checkDelay(self):
        """Checks and runs actions waiting in delay."""
        while True:
            again = False
            for delay in self.delay[:]:
                if delay['ticks'] <= 0:
                    self.delay.remove(delay)
                    delay['script'](delay['env'])
                    again = True

            if not again:
                break

        if self.logging:
            info("")

    def controlTime(self, time=None, script="tick"):
        """
        Increments internal time by 'time' ticks, runs tick script, and activates any delayed effects.

        .. code-block:: xml

            <i>TIME time</i>
        """
        if time is None:
            time = self.increment

        for i in range(time):
            # Increment time
            self.time += 1
            if self.logging:
                info(f'Env time is now {self.time}')

            # Process tick script
            if script in self:
                self.castMove(self["empty"], self["empty"], self[script])

            if self.delay:
                for delay in self.delay:
                    delay['ticks'] -= 1

                if self.logging:
                    info("Processing delayed effects")
                self._checkDelay()

            # Append new version to state.
            if self.state:
                self.state.append(self.copy())

        return self.cont

    def getRevert(self, time=0):
        """
        Returns a previous game state.

        Note: time represents how many ticks to revert.

        .. code-block:: python

            main_map.getRevert(5)
        """

        if self.state:
            new_states = self.state.copy()
            new = self
            for i in range(time + 1):
                try:
                    new = new_states.pop()
                except IndexError:
                    return False

            new.state = new_states
            new.state.append(new.copy())
            return new

    def controlDelay(self, callback, time, storage):
        """
        Delays an effect a number of ticks.

        Note: A delay of 0 will delay an action until the end of the current action window.
        Note: Notice the back tick.

        .. code-block:: xml

            <i>DELAY `(caster CAST target move) 0</i>
        """
        env = _getEnv(storage, {"self": self})

        new_delay = {
            "env": env,
            "script": self.getEntity(callback),
            "ticks": time
        }

        self.delay.append(new_delay)
        if self.logging:
            info(f"effect added to delay for {time}")
        return True

    @no_recurse
    def testFuture(self, caster, target, move, callback, time=0):
        """
        Experimental function. Checks if condition is true in the future.
        (Works for chess, I doubt I will touch it.)

        caster, target, move - Event to trigger.
        callback - Condition to check in the future.
        time - how much time to travel into the future.
        """
        if self.auto_entity:
            caster = self.getEntity(caster)
            target = self.getEntity(target)
            move = self.getEntity(move)

        # create test instance
        clone = self.copy()
        clone.revert = 0
        clone.logging = False

        # convert instance targets into clone targets.
        caster = clone.getEntity(caster.loc)
        target = clone.getEntity(target.loc)
        move = clone.getEntity(move.name)

        # Send clone into the future.
        env = clone.getEnv(caster, target, move)
        move.controlEffect(env)
        clone.controlTime(time)

        effects = clone.getEntity(callback)
        test = clone.testMove(caster, target, effects)

        return test

    # ---------------------------------------------------
    #           Corporate get functions                 -
    # ---------------------------------------------------

    def getStat(self, target, stat):
        """
        General getter for all attributes.

        Note: Attributes will include values inherited from their group.

        .. code-block:: xml

            <i>target [$caster] STAT stat</i>
        """
        if self.auto_entity:
            target = self.getEntity(target)
        if stat in {"increment", "group"}:
            return getattr(target, stat)
        elif stat == "envtime":
            return self.time
        elif stat not in self.lists.union(self.stats).union(self.strings):
            raise UndefinedSelectorError(stat, "stat")

        values = self._getStatHelper(target, stat)

        # rmap only apply to current object not rlist.
        if target.loc:
            for map_ in self.rmap:
                if map_ != target.loc[2]:
                    child = self[self.map[map_][target.loc]]
                    values = chain(values, self._getStatHelper(child, stat))

        if stat in self.stats:
            return sum(values)
        elif stat in self.lists:
            return [i for j in values for i in j]
        elif stat in self.strings:
            for i in values:
                if i != "":
                    return i
            else:
                return ""

    def _getStatHelper(self, target, stat):
        yield getattr(target, stat)
        for lst in self.rlist:
            for unit in getattr(target, lst):
                if isinstance(unit, str):
                    unit = self[unit]
                yield from self._getStatHelper(unit, stat)

    def getGroup(self, group: str):
        """Return all objects belonging to group.

         .. code-block:: xml

            <i>ALL group</i>
        """
        if self.auto_entity < 3:
            group = self.getEntity(group)

        return [unit for unit in self.values() if group.name in unit.group]

    def getListStat(self, target, lst, stat):
        """
        Returns sum of all object stats in a list.

        .. code-block:: xml

            <i>target [$caster] lst LISTSTAT stat</i>
        """
        lst = self.getStat(target, lst)

        found = (self.getStat(element, stat) for element in lst)

        if stat in self.stats:
            return sum(found)
        elif stat in self.lists:
            return list(i for j in found for i in j)
        elif stat in self.strings:
            return next(found)

    def getListGroup(self, target, lst, group):
        """
        Returns first element in list that is a member of group.

        .. code-block:: xml

            <i>target [$caster] lst LISTGROUP group</i>
            <i>target [$caster] CLASS group</i>
        """
        if self.auto_entity:
            group = self.getEntity(group)

        start = self.getStat(target, lst)
        for item in start:
            if group.name in self[item].group:
                return item
        return False

    def getSearch(self, *args):
        """
        Return objects contained in all given groups.

        .. code-block:: xml

            <i>SEARCH arg1 arg2 ...</i>
        """
        found = []
        for arg in args:
            for new in self.getGroup(arg):
                if new not in found:
                    found.append(new)

        return found

    # ---------------------------------------------------
    #                 Call wrappers                    -
    # ---------------------------------------------------

    def getEnv(self, caster, target, move):
        """Internal method for creating contexts for calls."""

        var_local = {
            "caster": caster,
            "target": target,
            "move": move,
            "C": caster,
            "T": target,
            "M": move,
        }
        return _getEnv(var_local, {"self": self})

    def _controlMove(self, casters, targets, moves):
        """
        Function to trigger an event. Returns True if an action triggers successfully.

        Note: If a list of entity selectors is received for any argument, all actions in the product will trigger.
        """
        if self.auto_entity:
            casters = self.getEntity(casters)
            targets = self.getEntity(targets)
            moves = self.getEntity(moves)

        value = False

        for caster, target, move in product(casters, targets, moves):
            if self.logging:
                # Logging
                caster_name = caster
                if caster.immutable and caster.loc:
                    caster_name = caster.loc

                target_name = target
                if target.immutable and target.loc:
                    target_name = target.loc

                if target == caster == self["empty"]:
                    info(f"init {move}")
                elif target == self["empty"]:
                    info(f"{caster_name} uses {move}")
                else:
                    info(f"{caster_name} uses {move} on {target_name}")

            # Cast the move
            env = self.getEnv(caster, target, move)
            for require in self.getStat(move, 'require'):
                try:
                    if not eval(require.code, env):
                        if self.logging:
                            info(f"failed because {require}")
                        break
                except Exception:
                    raise UdebsExecutionError(require)
            else:
                for effect in self.getStat(move, 'effect'):
                    try:
                        eval(effect.code, env)
                    except Exception:
                        raise UdebsExecutionError(effect)

                value = True

        return value

    def testMove(self, caster, target, move):
        """
        Simulates an action. Returns true if require passes successfully, False otherwise.

        .. code-block:: python

            main_map.testMove(caster, target, move)

        """
        if self.auto_entity:
            move = self.getEntity(move)
        env = self.getEnv(caster, target, move)
        return move.testRequire(env) is True

    def castInit(self, moves, times=1):
        """Cast an action without variables.

        .. code-block:: xml

            <i>INIT move</i>
        """
        for i in range(times):
            if not self.castMove(self["empty"], self["empty"], moves):
                return False

        return True


    def castAction(self, caster, move):
        """Cast an action with only a caster.

        .. code-block:: xml

            <i>caster [$caster] ACTION move</i>
        """
        return self.castMove(caster, self["empty"], move)

    def castMove(self, caster, target, move):
        """Cast an action including both a caster and an action.

        .. code-block:: xml

            <i>caster [$caster] CAST target move</i>
        """
        caster = self.getEntity(caster)
        target = self.getEntity(target)
        move = self.getEntity(move)

        value = self._controlMove(caster, target, move)
        if value:
            self._checkDelay()

        return value

    def castSingle(self, string):
        """
        Call a function directly from a user input effect String.

        .. code-block:: python

            main_map.castSingle("caster CAST target move")
        """
        code = Script(string, version=self.version)
        env = _getEnv({}, {"self": self})
        return code(env)

    # ---------------------------------------------------
    #               Entity control                     -
    # ---------------------------------------------------
    def controlListAdd(self, targets, lst, entries):
        """
        Adds an element to a list.

        .. code-block:: xml

            <i>caster [$caster] lst GETS entries</i>
        """
        if self.auto_entity:
            targets = self.getEntity(targets)

        if not isinstance(entries, list):
            entries = [entries]

        changed = False
        for target, entry in product(targets, entries):
            if not target.immutable:
                getattr(target, lst).append(entry)
                changed = True
                if self.logging:
                    info(f"{entry} added to {target} {lst}")

        return changed

    def controlListRemove(self, targets, lst, entries):
        """
        Removes an element from a list.

        .. code-block:: xml

            <i>caster [$caster] lst LOSES entries</i>
        """
        if self.auto_entity:
            targets = self.getEntity(targets)

        if not isinstance(entries, list):
            entries = [entries]

        changed = False
        for target, entry in product(targets, entries):
            if not target.immutable:
                value = getattr(target, lst)
                if entry in value:
                    changed = True
                    value.remove(entry)
                    if self.logging:
                        info(f"{entry} removed from {target} {lst}")

        return changed

    def controlClear(self, targets, lst):
        """
        Removes all from a targets list attribute.

        .. code-block:: xml

            <i>target [$caster] CLEAR lst</i>
        """
        if self.auto_entity:
            targets = self.getEntity(targets)

        changed = False
        for target in targets:
            if not target.immutable:
                getattr(target, lst).clear()
                changed = True
                if self.logging:
                    info(f"{target} {lst} has been cleared")

        return changed

    def controlShuffle(self, targets, lst):
        """
        Randomize order of a list.

        .. code-block:: xml

            <i>target SHUFFLE lst</i>
        """
        if self.auto_entity:
            targets = self.getEntity(targets)

        changed = False
        for target in targets:
            if not target.immutable:
                self.rand.shuffle(getattr(target, lst))
                changed = True
                if self.logging:
                    info(f"{target} {lst} has been shuffled")

        return changed

    def controlIncrement(self, targets, stat, increment, multi=1):
        """
        Increment a statistic by a static value.

        Note: Final value will be modified as follows (Stat = Stat + (increment * multi))

        .. code-block:: xml

            <i>target stat += increment</i>
            <i>target stat -= increment</i>
            <i>target [$caster] stat CHANGE increment</i>
        """
        if self.auto_entity:
            targets = self.getEntity(targets)

        changed = False
        total = int(increment * multi)
        for target in targets:
            if not target.immutable:
                setattr(target, stat, getattr(target, stat) + total)
                changed = True
                if self.logging:
                    info(f"{target} {stat} changed by {total} is now {self.getStat(target, stat)}")

        return changed

    def controlString(self, targets, stat, value):
        """
        Replace a value with another value.

        Note: This function also works on all attributes.

        .. code-block:: xml

            <i>caster [$caster] stat REPLACE value</i>
        """
        if self.auto_entity:
            targets = self.getEntity(targets)

        changed = False
        for target in targets:
            if not target.immutable:
                setattr(target, stat, value)
                changed = True
                if self.logging:
                    info(f"{target} {stat} changed to {value}")

        return changed

    def controlRecruit(self, target, positions):
        """
        Create a new entity by copying another then move it to a position.

        Note: If position is set to False copy will be made but not moved into a location.
        Note: One new entity will be created for each position given.

        .. code-block:: xml

            <i>caster [$caster] RECRUIT position</i>
        """
        if self.auto_entity:
            target = self.getEntity(target)

        new = target
        if not isinstance(positions, list):
            positions = [positions]

        for position in positions:
            if not target.immutable:
                new = target.clone()
                self[new.name] = new
                if self.logging:
                    info(f"{new} has been recruited")
            self.controlTravel(new, position)

        return new

    def controlDelete(self, target):
        """
        Permanently remove an object from the game.

        .. code-block:: xml

            <i>target DELETE</i>
        """
        if self.auto_entity:
            target = self.getEntity(target)

        if not target.immutable:
            del self[target.name]
            if target.loc:
                del self.map[target.loc[2]][target.loc]
            if self.logging:
                info(f"{target} has been deleted")
            return True
        return False

    # ---------------------------------------------------
    #                 Entity get wrappers              -
    # ---------------------------------------------------

    def getX(self, target, value=0):
        """
        Gets the x coordinate of target, False if target not on a map.

        .. code-block:: xml

            <i>target [$caster] XLOC</i>
        """
        loc = self.getLocObject(target)
        return loc[value] if loc else None

    def getY(self, target):
        """
        Gets the y coordinate of target, False if target not on a map.

        .. code-block:: xml

            <i>target [$caster] YLOC</i>
        """
        return self.getX(target, 1)

    def getMapName(self, target):
        """
        Gets the y coordinate of target, False if target not on a map.

        .. code-block:: xml

            <i>target [$caster] MAPNAME</i>
        """
        return self.getX(target, 2)

    def getLoc(self, target):
        """
        Gets the location tuple of a target, False if target not on a map.

        .. code-block:: xml

            <i>target [$caster] LOC</i>
        """
        return self.getLocObject(target)

    def getName(self, target):
        """
        Gets the name of a target.

        .. code-block:: xml

            <i>target [$caster] NAME</i>
        """
        if self.auto_entity:
            target = self.getEntity(target)

        names = []
        for i in target:
            names.append(i.name)

        return names if len(names) > 1 else names[0]

    def getShift(self, target, x, y, name=None):
        """
        Returns a new loc shifted x and y units from an old one

        .. code-block:: xml

            <i>target [$caster] SHIFT x y</i>
        """
        loc = self.getLocObject(target)
        if loc is not None:
            new_name = name if name else loc[2]
            new_loc = loc[0] + x, loc[1] + y, new_name
            if self.map[new_name].testLoc(new_loc):
                return new_loc

        return self['empty']

    # ---------------------------------------------------
    #                 Board get wrappers               -
    # ---------------------------------------------------

    def mapIter(self, name="map"):
        """Iterate over all cells in a map."""
        map_ = self.map[name]
        for loc in map_:
            yield self[map_[loc]]

    def getPath(self, caster, target, callback):
        """
        Finds a path between caster and target using callback as filter for valid space.

        .. code-block:: xml

            <i>PATH callback caster [$caster] target [$target]</i>
        """
        caster = self.getLocObject(caster)
        if caster:
            target = self.getLocObject(target)
            callback = self.getEntity(callback)
            map_ = self.map[caster[2]]
            if map_.testLoc(target):
                return map_.getPath(caster, target, callback=callback, state=self)

        return []

    def getDistance(self, caster, target, method):
        """
        Returns distance between caster and target using method as a metric.

        .. code-block:: xml

            <i>PATH method caster [$caster] target [$target]</i>
        """

        caster = self.getLocObject(caster)
        if caster:
            target = self.getLocObject(target)
            if method in self:
                method = self.getEntity(method)

            map_ = self.map[caster[2]]
            if map_.testLoc(target):
                return map_.getDistance(caster, target, method, state=self)

        return float("inf")

    def testBlock(self, caster, target, callback, max_dist=None):
        """
        Test to see if path exists between caster and target using callback as filter for valid space.

        .. code-block:: xml

            <i>BLOCK callback caster [$caster] target [$target]</i>
        """
        caster = self.getLocObject(caster)
        if caster:
            target = self.getLocObject(target)
            callback = self.getEntity(callback)

            map_ = self.map[caster[2]]
            if map_.testLoc(target):
                return map_.testBlock(caster, target, callback=callback, state=self, max_dist=max_dist)

        return False

    def getFill(self, center, callback=False, include_center=True, distance=float("inf")):
        """
        Gets all squares in a pattern starting at center, using callback as filter for valid space.

        .. code-block:: xml

            <i>FILL center callback include_center [true] distance [null]</i>
        """
        if distance is None:
            distance = float("inf")

        center = self.getLocObject(center)
        if center:
            callback = self.getEntity(callback)
            map_ = self.map[center[2]]
            fill = sorted(map_.getFill(center, distance, include_center,
                                       callback=callback, state=self))
            self.rand.shuffle(fill)
            return fill

        return []

    def printMap(self, board="map"):
        """Print a map."""
        self.map[board].show()

    # ---------------------------------------------------
    #              Board control wrappers              -
    # ---------------------------------------------------
    def controlTravel(self, caster, targets=None):
        """Move one object to a new location.

        .. code-block:: xml

            <i>caster [$caster] MOVE target</i>
        """
        if self.auto_entity:
            caster = self.getEntity(caster)

        if not isinstance(targets, list):
            targets = [targets]

        if len(targets) > 1 and not caster.immutable:
            raise Exception("Non immutable entities cannot be moved to multiple locations.")

        for target in targets:
            # First remove caster from it's current location
            if not caster.immutable and caster.loc:
                del self.map[caster.loc[2]][caster.loc]

            # Then move caster to target location.
            target = self.getLocObject(target)
            if target:
                map_ = self.map[target[2]]
                unit_name = map_[target]
                unit_self = self[unit_name]

                map_[target] = caster.name
                if not unit_self.immutable:
                    unit_self.loc = None

                if self.logging:
                    info(f"{caster} has moved to {target}")

            # Update the entity itself.
            if not caster.immutable:
                caster.loc = target

    # ---------------------------------------------------
    #                 Game Loop Helpers                -
    # ---------------------------------------------------

    def gameLoop(self, time=1, script="tick"):
        """Iterate over in game time.

        .. code-block:: python

            for state in main_map.gameLoop():
                state.castMove(caster, target, move)
        """
        current = self

        current.increment = time
        while current.cont:
            yield current

            if current.next is not None:
                yield current.next
                current = current.next

            current.controlTime(current.increment, script=script)

        if current.logging:
            info(f"EXITING {current.name}\n")

    def exit(self, value=1):
        """Requests the end of the game by setting Instance.cont to False, and exiting out of the game_loop.

        .. code-block:: xml

            <i>EXIT</i>
        """
        if self.logging:
            info("Exit requested with value of: {}".format(value))
        self.cont = False
        self.value = value

    def resetState(self, script="reset"):
        """Resets the game metadata to default.

        .. code-block:: python

            main_map.resetState()
        """
        self.cont = True
        self.value = None
        self.time = 0
        self.delay.clear()
        if self.logging:
            info(f"Env time is now {self.time}")

        if script in self:
            self.castMove(self["empty"], self["empty"], self[script])
        elif self.logging:
            info("")

        if self.state:
            self.state.clear()
            self.state.append(self.copy())



import functools
import math
import time
from .instance import Instance
from .utilities import cache



class State(Instance):
    """Base class for all tree search algorithms."""
    def substates(self, override=None):
        """Iterate over substates to a state.

        yields a tuple, (child instance, action tuple)
        if the legalMoves iterator yielded something other than a tuple. This method will yield just that object for both values.
        """
        stateNew = None
        moves = override
        if moves is None:
            moves = self.legalMoves()

        for move in moves:
            if isinstance(move, tuple):
                if stateNew is None:
                    stateNew = self.copy()

                if stateNew.castMove(*move[:3]):
                    stateNew.controlTime(self.increment)
                    yield stateNew, move
                    stateNew = None
            else:
                yield move, move

    def hash(self):
        """Implement this function if you plan on using treesearch.cache"""
        raise NotImplementedError

    def legalMoves(self):
        """This function must be implemented. Iterate over all (caster, target, move) tuples that are valid children of current node."""
        raise NotImplementedError

    def endState(self):
        """Example implementations use this function to fetch a states leaf value."""
        return self.value

#---------------------------------------------------
#             Various Tree Search Algorithms       -
#---------------------------------------------------
class BruteForce(State):
    """Example implementation of a brute force minimax solver."""
    @cache
    def result(self):
        value = self.endState()
        if value is not None:
            return -abs(value)

        results = []
        for child, e in self.substates():
            if child is not e:
                child = child.result()

            results.append(-child)

        return max(results)

class AlphaBeta(State):
    """Example implementation of an alphabeta minimax solver."""
    @cache
    def result(self, alpha=-float("inf"), beta=float("inf")):
        value = self.endState()
        if value is not None:
            return -abs(value)

        for child, e in self.substates():
            if child is not e:
                child = child.result(-beta, -alpha)

            result = -child

            if result > alpha:
                alpha = result

            if alpha >= beta:
                break

        return alpha

class MarkovChain(State):
    """Example implementation of a markov chain calculator."""
    @cache
    def result(self):
        value = self.endState()
        if value is not None:
            return value

        total = 0
        for child, p in self.substates():
            total += child.result() * p[3]

        return total

#---------------------------------------------------
#                   Monty Carlo                    -
#---------------------------------------------------
class MontyCarlo(State):
    """Experimental: Example implementation of a Monty Carlo treesearch algorithm."""
    def init(self, entry=None):
        self.children = None
        self.v = 0
        self.n = 0
        self.entry = entry
        return self

    def result(self, maximizer=True, **kwargs):
        if self.children is None:
            value, *args = self.simulate(maximizer, **kwargs)
            self.children = [c.init(e, *args) for c, e in self.substates()]

        else:
            # selection
            node = max(self.children, key=self.weight)
            value = node.result(not maximizer, **kwargs)
            value = 1 - value

        self.v += value
        self.n += 1
        return self.v / self.n

    def weight(self, child, c=1):
        if child.n == 0:
            return float("inf")

        q = 1 - (child.v / child.n)
        un = math.sqrt(math.log(self.n) / child.n)
        return q + (c * un)

    def think(self, think_time, verbose=True, **kwargs):
        t_end = time.time() + think_time
        iters = 0
        while time.time() < t_end:
            self.result(**kwargs)
            iters +=1

        if verbose:
            print(f"processed {iters} new iterations")

    def select(self, f=None):
        if f is None:
            f = (lambda x: x.n)

        return max(self.children, key=f)

    def simulate(self, maximizer=True):
        raise NotImplementedError

class AlphaMontyCarlo(MontyCarlo):
    """Experimental: Alpha Zero inspired variant of MontyCarlo treesearch."""
    def init(self, entry=None, policy=None):
        self.prior = None if policy is None else policy.get(entry[1], 0)
        return super().init(entry)

    def weight(self, child, c=1, p=0.5):
        q = 1 - (child.v / child.n if child.n > 0 else p)
        un = math.sqrt(self.n / (child.n + 1))
        u = c * child.prior * un
        return q + u



from udebs.interpret import *
from pytest import raises
import udebs


class TestInterpret:
    @staticmethod
    def setup():
        test = {
            "solitary": {
                "f": "solitary",
            },
            "testing": {
                "f": "TEST",
                "default": {"$3": 50},
                "args": ["-$1", "$1", "$2", "three"],
                "kwargs": {"none": "$3", "value": "empty", "test": 10},
            }
        }
        importModule(test, {'self': None})

    def test_setup(self):
        assert "solitary" in Variables.keywords()
        assert "testing" in Variables.keywords()
        assert "self" in Variables.env

    def test_list(self):
        assert interpret("one two three") == "('one','two','three')"
        assert interpret("1 2 three") == "(1,2,'three')"

    def test_Syntax_mismatched_brackets(self):
        with raises(UdebsSyntaxError):
            interpret("(one two (three)")

    def test_Syntax_unused_arguments(self):
        with raises(UdebsSyntaxError):
            interpret("one in two unused, unused2")

    def test_Syntax_empty_callstring(self):
        with raises(UdebsSyntaxError):
            interpret("() == one")

    def test_Syntax_extra_keywords(self):
        with raises(UdebsSyntaxError):
            interpret("max one in two")

    def test_solitary(self):
        assert interpret("one in (solitary)") == "standard.inside('one',solitary(),1)"

    def test_dot(self):
        assert interpret("one two one.two.three") == interpret("one two (one two three)")

    def test_prefix(self):
        assert interpret("one two -three") == "('one','two',operator.sub(0,'three'))"
        assert interpret("-one.two.three") == "(operator.sub(0,'one'),'two','three')"
        assert interpret("-(one two three)") == "operator.sub(0,('one','two','three'))"

    def test_call(self):
        assert interpret(
            "negative testing one two") == "TEST('negative','one','two','three',none=50,test=10,value='empty')"

    def test_global(self):
        assert interpret("one $target") == "('one',storage.__getitem__('target'))"

    def test_specials(self):
        assert interpret("self") == "(self)"
        assert interpret("true") == "(True)"
        assert interpret("'help'") == "('help')"

    def test_debug(self):
        interpret("(one) two three", debug=True)

    def test_Error_repr(self):
        one = UdebsSyntaxError("test")
        print(one)


class TestBase:
    def setup(self):
        path = os.path.dirname(__file__)
        self.env = udebs.battleStart(path + "/test.xml", log=True)

    def test_logicif(self):
        assert self.env.castSingle("if 1 0 2") == 0
        assert self.env.castSingle("if 0 0 2") == 2

    def test_logicor(self):
        assert self.env.castSingle("1 or 0") is True
        assert self.env.castSingle("0 or 0") is False
        assert self.env.castSingle("0 or `move1") is True

    def test_var(self):
        # not sure how to test if this actually worked
        assert self.env.castSingle("1 = test")
        assert self.env.castSingle("1 -> test")

    def test_inside(self):
        assert self.env.castSingle("1 in (1 0)") is True
        assert self.env.castSingle("1 in (1 0) 2") is False
        assert self.env.castSingle("1 in (2 0) 0") is True

        assert self.env.castSingle("a in alberta") is True

    def test_notin(self):
        assert self.env.castSingle("1 not-in (0 2)") is True

    def test_equal(self):
        assert self.env.castSingle("1 == 1 1 1 1") is True
        assert self.env.castSingle("1 == 1 2 1 1") is False

    def test_not_equal(self):
        assert self.env.castSingle("1 != 1 2 3 4") is False
        assert self.env.castSingle("1 != 2 3 4 5") is True
        assert self.env.castSingle("1 != 2") is True

    def test_relations(self):
        assert self.env.castSingle("1 > 0") is True
        assert self.env.castSingle("1 < 0") is False
        assert self.env.castSingle("1 <= 1") is True
        assert self.env.castSingle("0 >= 1") is False

    def test_math(self):
        assert self.env.castSingle("+ 1 2 3 4 5") == 15
        assert self.env.castSingle("* 1 2 3") == 6
        assert self.env.castSingle("5 % 10") == 5
        assert self.env.castSingle("10 / 5") == 2

    def test_bool(self):
        assert self.env.castSingle("!true") is False

    def test_element(self):
        assert self.env.castSingle("(1 2 3) elem 0") == 1

    def test_len(self):
        assert self.env.castSingle("length (0 0 0 1)") == 4



from udebs import interpret
import udebs
from pytest import raises
import os


class TestBattleStart:
    def test_alternate(self):
        tests = [["one", "two"], "three", ["four", "five", "six"]]
        test = list(udebs.alternate(*tests))
        assert len(test) == 3

    def test_timer(self):
        with udebs.Timer() as f:
            for i in range(3):
                assert i > -1

        assert f.time > 0

    def test_placeholder(self):
        """Not function is deprecated"""
        udebs.placeholder("TEST")
        assert "TEST" in interpret.Variables.modules[-1]

    def test_lookup(self):
        udebs.lookup("LOOKUP", {"one": {"two": {"three": 4}}})
        path = os.path.dirname(__file__)
        env = udebs.battleStart(path + "/test.xml", log=True)
        assert env.castSingle("LOOKUP one two three") == 4
        test = env.castSingle("LOOKUP one three")
        assert test == 0

    def test_register(self):

        @udebs.register({"args": ["$1", "$2"]})
        def ADDITION1(one, two):
            return one + two

        def ADDITION2(one, two):
            return one + two

        udebs.register(ADDITION2, ["$1", "$2"])

        @udebs.register(["$1"])
        class SUBTRACTION:
            def __init__(self):
                self.one = 7

            def __call__(self, two):
                return self.one - two

        udebs.register(SUBTRACTION, ["$1"], name="sub1")

        main_map = udebs.battleStart()
        assert main_map.castSingle("ADDITION1 1 1") == 2
        assert main_map.castSingle("ADDITION2 1 3") == 4
        assert main_map.castSingle("SUBTRACTION 4") == 3
        assert main_map.castSingle("sub1 4") == 3



from collections.abc import MutableMapping
import math
import copy

sides = [
    (-1, 0),
    (0, 1),
    (1, 0),
    (0, -1),
    (-1, 1),
    (1, -1),
    (1, 1),
    (-1, -1),
]


class Board(MutableMapping):
    """This is a test."""

    def __init__(self, **options):
        if "_data" in options:
            # second path for copy operation only
            self.__dict__ = options["_data"]
            return

        self.name = options.get("name", "map")
        self.empty = options.get("empty", "empty")
        self.type = options.get("type", False)
        self.count = 4
        if self.type:
            self.count += 2
            if self.type == "diag":
                self.count += 2

        # Set up maps.
        dim = options.get("dim", (1, 1))
        if isinstance(dim, tuple):
            self.map = []
            for e in range(dim[0]):
                self.map.append([self.empty for _ in range(dim[1])])

        elif isinstance(dim, list):
            self.map = dim

        self.x = len(self.map)
        self.y = max([len(x) for x in self.map])

    def __eq__(self, other):
        if not isinstance(other, Board):
            return False

        if self.map == other.map:
            if self.name == other.name:
                if self.empty == other.empty:
                    return True
        return False

    def __getitem__(self, key):
        x, y = key[0], key[1]
        if x >= 0:
            if y >= 0:
                return self.map[x][y]
        raise IndexError

    def __setitem__(self, key, value):
        x, y = key[0], key[1]
        if x >= 0:
            if y >= 0:
                self.map[x][y] = value
                return
        raise IndexError

    def __delitem__(self, key):
        x, y = key[0], key[1]
        if x >= 0:
            if y >= 0:
                self.map[x][y] = self.empty
                return
        raise IndexError

    def __len__(self):
        return self.x * self.y

    def __iter__(self):
        for x in range(self.x):
            for y in range(self.y):
                yield x, y, self.name

    def __repr__(self):
        return "<board: " + self.name + ">"

    def __copy__(self):
        return self.copy()

    def copy(self):
        options = {}
        for k, v in self.__dict__.items():
            if k == "map":
                v = [i[:] for i in self.map]
            options[k] = v

        return Board(_data=options)

    # ---------------------------------------------------
    #                    Methods                       -
    # ---------------------------------------------------
    def show(self):
        """Pretty prints a map."""
        maxi = 0
        for name in self.values():
            if name != self.empty and len(str(name)) > maxi:
                maxi = len(str(name))

        for y in range(self.y):
            if self.type == "hex":
                print(" " * y * maxi, end="")

            for x in range(self.x):
                value = self[x, y]
                if value == self.empty:
                    value = "_"

                print(str(value).ljust(maxi), end=" ")

            print()

    def getDistance(self, one, two, method, **kwargs):
        """Returns distance between two locations.

        one - starting loc
        two - end loc
        method - metric or callback to use

        """
        if not isinstance(method, str):
            count = 0
            for i in self.getAdjacent(one, callback=method, **kwargs):
                if two in i:
                    return count
                count += 1
            return float("inf")
        elif method == 'y':
            return int(abs(one[1] - two[1]))
        elif method == 'x':
            return int(abs(one[0] - two[0]))
        elif method == 'z':
            return int(abs(two[0] + two[1]) - (one[0] + one[1]))
        elif method == 'hex':
            return int((abs(one[0] - two[0]) + abs(one[1] - two[1]) + abs((two[0] + two[1]) - (one[0] + one[1]))) / 2)
        elif method == "p1":
            return int(abs(one[0] - two[0]) + abs(one[1] - two[1]))
        elif method == "p2":
            return int(math.sqrt(math.pow(one[0] - two[0], 2) + math.pow(one[1] - two[1], 2)))
        elif method == "pinf":
            return int(max(abs(one[0] - two[0]), abs(one[1] - two[1])))
        else:
            raise ValueError(f"{method} is an invalid metric")

    def getAdjacent(self, start, sort=None, pointer=None, state=None, callback=None):
        """
        Iterates over concentric circles of adjacent tiles.

        new - starting local
        sort - a function that can be used to sort the order nodes are explored.
        pointer - optional dictionary for tracking pathing
        state - the state if callback is to be used
        callback - callback to filter cells

        """
        if isinstance(start, set):
            new = start
            start = next(iter(start))
        else:
            new = {start}

        searched = copy.copy(new)

        while len(new) > 0:
            yield new

            next_ = set()
            for child in new:
                for x_, y_ in sides[:self.count]:
                    loc = (child[0] + x_, child[1] + y_, self.name)
                    if loc not in searched:
                        searched.add(loc)
                        if self.testLoc(loc):
                            if callback:
                                env = state.getEnv(start, loc, callback)
                                if not callback.testRequire(env) is True:
                                    continue
                            next_.add(loc)
                            if pointer:
                                pointer[loc] = child
            new = next_
            if sort:
                new = sort(new)

    def getPath(self, start, finish, **kwargs):
        """
        Algorithm to find a path between start and finish assuming callback.

        start, finish - Start and finish locations.

        Returns empty list if there is no path.

        """
        if not isinstance(start, set):
            start = {start}

        if not isinstance(finish, set):
            finish = {finish}

        pointer = {i: None for i in start}

        # Populate pointer.
        for i in self.getAdjacent(start, pointer=pointer, **kwargs):
            final = finish.intersection(i)
            if len(final) > 0:
                break
        else:
            return []

        new = final.pop()
        found = [new]
        while pointer[new]:
            new = pointer[new]
            found.append(new)

        found.reverse()
        return found

    def getFill(self, center, distance=float("inf"), include_center=True, **kwargs):
        """
        Returns a list of map spaces that center fills into.

        center - Starting loc
        distance - Maximum fill distance from target.
        include_center - Boolean to determine if center should be included.

        """
        if not isinstance(center, set):
            center = {center}

        found = set()
        count = 0
        for i in self.getAdjacent(center, **kwargs):
            found.update(i)
            if count >= distance:
                break
            count += 1

        if not include_center:
            found -= center

        return found

    def testLoc(self, loc):
        """
        Test a loc to see if it is valid.

        loc - Starting loc

        """
        if loc:
            try:
                map_ = loc[2]
            except IndexError:
                map_ = "map"

            if self.name == map_:
                if 0 <= loc[0] < self.x:
                    if 0 <= loc[1] < self.y:
                        return True

        return False

    def testBlock(self, start, finish, max_dist=None, **kwargs):
        """
        Tests to see if there exists a path from start to finish.

        start, finish - Starting and finishing locations.

        """
        if max_dist is None:
            max_dist = float("inf")

        if start == finish and max_dist > 0:
            return True

        i = self.getAdjacent(finish)
        next(i)
        final = next(i)
        for dist, i in enumerate(self.getAdjacent(start, **kwargs)):
            if i.intersection(final):
                return True

            if dist + 1 >= max_dist:
                return False

        return False



class Entity:
    def __init__(self, field=None, **options):
        if "_data" in options:
            # second path for copy operation only
            self.__dict__ = options["_data"]
            return

        self.name = options.get("name", "")
        self.loc = options.get("loc", None)
        self.immutable = options.get("immutable", False)
        self.increment = options.get("increment", 0)

        for stat in field.stats:
            self.__dict__[stat] = options.get(stat, 0)
        for lists in field.lists:
            self.__dict__[lists] = options.get(lists, [])
        for string in field.strings:
            self.__dict__[string] = options.get(string, '')

    def __eq__(self, other):
        if not isinstance(other, Entity):
            return False

        return self.__dict__ == other.__dict__

    def __len__(self):
        return 1

    def __repr__(self):
        return f"<entity: {self.name}{'!' if self.immutable else ''}>"

    def __iter__(self):
        yield self

    def __str__(self):
        return self.name

    # ---------------------------------------------------
    #                 Call Functions                    -
    # ---------------------------------------------------
    def controlEffect(self, env):
        for effect in env["self"].getStat(self, 'effect'):
            effect(env)

        return True

    def testRequire(self, env):
        for require in env["self"].getStat(self, 'require'):
            if not require(env):
                return require

        return True

    def __call__(self, env):
        value = self.testRequire(env)
        if value is True:
            return self.controlEffect(env)
        return value

    # ---------------------------------------------------
    #                Clone Functions                   -
    # ---------------------------------------------------
    def copy(self, **kwargs):
        """The dependency on field prevents me from doing this as __copy__"""

        for k, v in self.__dict__.items():
            if k not in kwargs:
                if isinstance(v, list):
                    v = v[:]
                kwargs[k] = v

        return Entity(_data=kwargs)

    def clone(self):
        """Returns a clone of self."""
        # Set name of new
        self.increment += 1
        name = self.name + str(self.increment)

        # Create new
        return self.copy(name=name, increment=0)



import copy
import re
import sys
import logging
from . import board, entity, instance, interpret
from xml.etree import ElementTree
from collections import deque


# This is a pretty printing algorithm I stole from the internet.
def _indent(elem, level=0):
    """A simple pretty print function for xml."""
    i = "\n" + level * "  "
    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = i + "  "
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
        for elem in elem:
            _indent(elem, level + 1)
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = i


# creates an xml file from instance object.
def battleWrite(env, location, pretty=False):
    """
    Writes an instance object to file.

    env - Instance object to write.
    location - Variable to pass to xml.etree.write.
    """
    e = ElementTree
    root = e.Element('udebs')

    def add_leaf(root2, node2, value2):
        new = e.SubElement(root2, node2)
        new.text = value2

    # Definitions
    definitions = e.SubElement(root, 'definitions')
    for dtype in ['stats', 'lists', 'strings']:
        middle = e.SubElement(definitions, dtype)
        for item in getattr(env, dtype):
            if item in {'group', 'effect', 'require', 'increment'}:
                continue
            final = e.SubElement(middle, item)
            if item in env.rlist:
                final.attrib['rlist'] = ''

    # Config variables.
    config = e.SubElement(root, 'config')
    if env.name != 'Unknown':
        add_leaf(config, "name", env.name)
    if env.logging is not True:
        add_leaf(config, "logging", str(env.logging))
    if env.revert != 0:
        add_leaf(config, "revert", str(env.revert))
    if env.version != 1:
        add_leaf(config, "version", str(env.version))
    if env.seed is not None:
        add_leaf(config, "seed", str(env.seed))
    if env.immutable is not True:
        add_leaf(config, "immutable", str(env.immutable))
    if env.auto_entity is not True:
        add_leaf(config, "auto_entity", str(env.auto_entity))

    # Time variables
    var = e.SubElement(root, 'var')
    if env.time != 0:
        add_leaf(var, "time", str(env.time))
    if env.increment != 1:
        add_leaf(var, "increment", str(env.increment))
    if env.cont is not True:
        add_leaf(var, "cont", str(env.cont))
    if env.value is not None:
        add_leaf(var, "value", str(env.value))

    # map
    maps = e.SubElement(root, 'maps')
    for map_ in env.map.values():
        node = e.SubElement(maps, map_.name)
        if map_.name in env.rmap:
            node.attrib['rmap'] = ''
        if map_.empty != 'empty':
            node.attrib['empty'] = map_.empty
        if map_.type is not False:
            node.attrib['type'] = map_.type
        for row in (list(i) for i in zip(*map_.map)):
            add_leaf(node, "row", ", ".join(row))

    # entities
    stats = env.stats.union(env.strings, env.lists)
    entities = e.SubElement(root, 'entities')
    special = e.SubElement(root, 'special')
    for item in env.values():
        if item.name == 'empty':
            continue

        elif isinstance(item.name, interpret.UdebsStr):
            add_leaf(special, "i", str(item.name))
            continue

        entity_node = e.SubElement(entities, item.name)

        if item.immutable:
            entity_node.attrib['immutable'] = ''

        for stat in stats:
            value = getattr(item, stat)

            if value in [0, '', []]:
                continue

            stat_node = e.SubElement(entity_node, stat)
            if stat not in env.lists:
                stat_node.text = str(value)
            elif len(value) == 1:
                stat_node.text = str(value[0])
            else:
                for elem in value:
                    add_leaf(stat_node, "i", str(elem))

    # Delay
    delay = e.SubElement(root, "delays")
    if delay is not None:
        for ent in env.delay:
            d = e.SubElement(delay, "delay")
            add_leaf(d, "ticks", str(ent["ticks"]))
            add_leaf(d, "script", str(ent["script"]))

            storage = e.SubElement(d, "storage")
            for key, value in ent["env"]["storage"].items():
                add_leaf(storage, key, str(value))

    # rand
    add_leaf(root, "random", str(env.rand.getstate()))

    # Final Cleanup
    if pretty:
        _indent(root)

    for node in root[:]:
        if node.text is None:
            root.remove(node)

    e.ElementTree(root).write(location)
    return True


# Creates and instance object from xml file.
def battleStart(xml_file=None, debug=False, script="init", name=None, revert=None, log=None, version=None, seed=None,
                immutable=None, field=None, auto_entity=None):
    """
    Creates an instance object from given xml file.

    xml_file - String representing file to look in.
    debug - Boolean that gets passed to the interpret.interpret function.
    """
    if xml_file is None:
        xml_file = "<udebs />"

    try:
        tree = ElementTree.parse(xml_file)
        root = tree.getroot()
    except IOError:
        root = ElementTree.fromstring(xml_file)

    # ENV
    if field is None:
        field = instance.Instance()

    # Definition
    defs = root.find("definitions")
    if defs is not None:
        for i in ("stats", "strings"):
            def_stats = defs.find(i)
            if def_stats is not None:
                for stat in def_stats:
                    getattr(field, i).add(stat.tag)

        def_lists = defs.find("lists")
        if def_lists is not None:
            for stat in def_lists:
                field.lists.add(stat.tag)
                if stat.get('rlist') is not None:
                    field.rlist.append(stat.tag)

    def fill_simple(root2, stat2, f, overwrite=None):
        if overwrite is not None:
            setattr(field, stat2, overwrite)
        else:
            tmp = root2.findtext(stat2)
            if tmp is not None:
                if f is not None:
                    tmp = f(tmp)

                setattr(field, stat2, tmp)

    # Config
    config = root.find("config")
    if config is not None:
        fill_simple(config, "name", None, name)
        fill_simple(config, "revert", int, revert)
        fill_simple(config, "logging", eval, log)
        fill_simple(config, "version", int, version)
        fill_simple(config, "seed", int, seed)
        fill_simple(config, "immutable", eval, immutable)
        fill_simple(config, "auto_entity", eval, auto_entity)

    # Time variables
    time = root.find("var")
    if time is not None:
        fill_simple(time, "time", int)
        fill_simple(time, "increment", int)
        fill_simple(time, "cont", eval)
        fill_simple(time, "next", eval)
        fill_simple(time, "value", eval)

    # Maps
    def addMap(field_map2):
        options2 = {"name": field_map2.tag}

        # Attributes
        for att in ("empty", "rmap", "type"):
            if field_map2.get(att) is not None:
                options2[att] = field_map2.get(att)

        # dimensions.
        dim_map = field_map2.find("dim")
        if dim_map is not None:
            options2['dim'] = (
                int(dim_map.find('x').text),
                int(dim_map.find('y').text)
            )

        else:
            dim = []
            for row in field_map2:
                dim.append(re.split(r"\W*,\W*", row.text))
            options2['dim'] = [list(j) for j in zip(*dim)]

        # Add to field
        field.map[options2["name"]] = board.Board(**options2)
        if "rmap" in options2:
            field.rmap.append(options2["name"])

    field_maps = root.find("maps")
    if field_maps is not None:
        [addMap(i) for i in field_maps]

    field_map = root.find("map")
    if field_map is not None:
        addMap(field_map)

    # Entities
    field["empty"] = entity.Entity(field, name="empty", immutable=True)

    # Create all entity type objects.
    entities = root.find("entities")
    if entities is not None:
        for item in entities:
            options = {"name": item.tag}

            immutable = item.get("immutable")
            if immutable is None:
                options["immutable"] = field.immutable
            elif immutable == "False":
                options["immutable"] = False
            else:
                options["immutable"] = True

            for stat in field.stats:
                elem = item.find(stat)
                if elem is not None:
                    options[stat] = int(elem.text)

            for string in field.strings:
                value = item.findtext(string)
                if value is not None:
                    options[string] = value

            for lst in field.lists:
                new_list = []
                find_list = item.find(lst)
                if find_list is not None:
                    if len(find_list) == 0:
                        find_list = [find_list]
                    for value in find_list:
                        if lst in {"effect", "require"}:
                            value.text = interpret.Script(value.text, field.version, debug)
                        new_list.append(value.text)
                options[lst] = new_list

            # set loc
            if not options["immutable"]:
                for map_ in field.map.values():
                    for loc in map_:
                        if map_[loc] == options["name"]:
                            options["loc"] = loc
                            break

            field[options["name"]] = entity.Entity(field, **options)

    # Special
    special = root.find("special")
    if special is not None:
        for i in special:
            field.getEntity(interpret.UdebsStr(i.text))

    # Delay
    delays = root.find("delays")
    if delays is not None:
        for delay in delays:
            callback = delay.findtext("script")
            time = int(delay.findtext("ticks"))
            storage = {node.tag: field[node.text] for node in delay.find("storage")}
            field.controlDelay(callback, time, storage)

    # Random
    rand = root.find("random")
    if rand is not None:
        field.rand.setstate(eval(rand.text))
    elif field.seed:
        field.rand.seed(field.seed)

    # Final cleanup
    logging.basicConfig(**{
        "stream": sys.stdout,
        "level": logging.INFO if field.logging else logging.WARNING,
        "format": "%(message)s",
    })

    if field.logging:
        logging.info(f"INITIALIZING {field.name}")
        logging.info(f"Env time is now {field.time}")

    if script in field:
        field.castInit(script)
    elif field.logging:
        logging.info("")

    if field.revert:
        field.state = deque(maxlen=field.revert + 1)
        field.state.append(copy.copy(field))

    return field



import udebs
import os
from udebs.board import Board


# Test battleStart
class TestBattleStart:
    def setup(self):
        path = os.path.dirname(__file__)
        self.test = udebs.battleStart(path + "/test.xml")

    def test_config(self):
        assert self.test.logging is False
        assert self.test.name == "testing"
        assert self.test.revert == 4
        assert self.test.version == 1
        assert self.test.seed == 690
        assert self.test.immutable is False

    def test_definitions(self):
        assert "ACT" in self.test.stats
        assert "DESC" in self.test.strings
        assert "equipment" in self.test.lists
        assert "inventory" in self.test.rlist
        total = sum([len(i) for i in [self.test.lists, self.test.stats, self.test.strings]])
        assert total == 7

    def test_map(self):
        assert len(self.test.map) == 3
        assert isinstance(self.test.map['two'], Board)
        assert self.test.map['map'].x == 5
        assert self.test.map['map'].y == 6

        two = self.test.map['two']
        assert two.empty == 'immune'
        assert two.name == 'two'
        assert two.y == 3
        assert two.x == 2

    def test_loc_added(self):
        assert self.test["unit1"]

    def test_var(self):
        assert self.test.time == 4
        # delay not tested yet.

    def test_entity(self):
        # four defined plus one built in.
        assert len(self.test) == 13
        assert len(self.test["unit1"].group) == 1

        stats = self.test.strings.union(self.test.stats, self.test.lists)
        for value in self.test.values():
            for stat in stats:
                assert hasattr(value, stat)

    def test_string(self):
        test = udebs.battleStart("<udebs><config><logging>True</logging></config><map><x>6</x><y>6</y></map></udebs>")
        assert len(test) == 1


class TestBattleWrite:
    def test_equal(self):
        path = os.path.dirname(__file__) + "/test.xml"
        path2 = os.path.dirname(__file__) + "/write_test.xml"

        env1 = udebs.battleStart(path, name="test")
        udebs.battleWrite(env1, path2, True)
        env2 = udebs.battleStart(path2, script=None)
        assert env1 == env2
        os.remove(path2)



import udebs
import os
from pytest import raises
import copy


class TestBoardClass:
    def setup(self):
        path = os.path.dirname(__file__)
        self.test = udebs.battleStart(path + "/test.xml")
        self.one = self.test.map["one"]
        self.two = self.test.map["two"]
        self.map = self.test.map['map']

    def test_size(self):
        assert self.one.y == 4
        assert self.two.x == 2
        assert len(self.map) == 30

    def test_get(self):
        assert "unit1" == self.two[(0, 0)]

    def test_get2(self):
        with raises(IndexError):
            self.two[(3, 3)]
        with raises(IndexError):
            self.two[(3, 3)] = "test"
        with raises(IndexError):
            del self.two[(3, 3)]

    def test_get3(self):
        with raises(IndexError):
            self.one[(-1, 0)]
        with raises(IndexError):
            self.one[(-1, 0)] = "test"
        with raises(IndexError):
            del self.one[(-1, 0)]

    def test_get4(self):
        with raises(TypeError):
            self.one[False]
        with raises(TypeError):
            self.one[False] = "test"
        with raises(TypeError):
            del self.one[False]

    def test_insert(self):
        self.one[(0, 0)] = "unit1"
        assert self.one[(0, 0)] == "unit1"
        assert self.one[(1, 0)] == "empty"

    def test_len(self):
        assert len(self.two) == 6
        assert len(self.one) == 16

    def test_iter(self):
        assert len(list(self.two)) == 6
        assert len(list(self.one)) == 16

    def test_x_y(self):
        assert self.two.x == 2
        assert self.two.y == 3

    def test_copy(self):
        three = copy.copy(self.two)
        assert three == self.two
        self.two[(0, 0)] = "empty"
        assert three != self.two

    def test_del(self):
        assert self.two[0, 0] != "empty"
        del self.two[0, 0]
        assert self.two[0, 0] == "immune"

    def test_eq(self):
        assert self.two != "two"

    def test_show(self):
        self.two.show()

    def test_repr(self):
        assert repr(self.two) == "<board: two>"


class TestPathing:
    def setup(self):
        path = os.path.dirname(__file__)
        self.test = udebs.battleStart(path + "/test.xml")
        self.test.controlTravel("unit1", (0, 0))
        self.test.controlTravel("unit2", (4, 3))
        self.test.controlTravel("immune", (2, 2))

    def test_getPath(self):
        test = self.test.getPath("unit1", "unit2", "notempty")
        assert (0, 0, "map") in test
        assert (4, 3, "map") in test
        assert len(test) == 5
        test = self.test.getPath((0, 1), (0, 1), "empty")
        assert len(test) == 1
        self.test.controlTravel("unit1", (1, 2, "two"))
        assert self.test.getPath((0, 0, "two"), (1, 2, "two"), "notempty") == []
        assert len(self.test.getPath("empty", "empty", "empty")) == 0
        assert self.test.getPath("unit1", "empty", "empty") == []

    def test_distance(self):
        assert self.test.getDistance("empty", "unit1", "x") == float("inf")
        assert self.test.getDistance("unit1", "empty", "x") == float("inf")
        assert self.test.getDistance("unit1", "empty", "empty") == float("inf")
        assert self.test.getDistance("unit1", "unit2", 'x') == 4
        assert self.test.getDistance("unit1", "unit2", 'y') == 3
        assert self.test.getDistance("unit1", "unit2", "z") == 7
        assert self.test.getDistance("unit1", "unit2", "hex") == 7
        assert self.test.getDistance("unit1", "unit2", 'p1') == 7
        assert self.test.getDistance("unit1", "unit2", "p2") == 5
        assert self.test.getDistance("unit1", "unit2", 'pinf') == 4
        assert self.test.getDistance("unit1", "unit2", "notempty") == 4

        with raises(ValueError):
            self.test.getDistance("unit1", "unit2", "not a unit")

    def test_testBlock(self):
        assert self.test.testBlock("empty", "unit1", "empty") is False
        assert self.test.testBlock("unit1", "unit2", "notempty")
        assert self.test.testBlock("unit1", "empty", "notempty") is False

    def test_getFill(self):
        test = self.test.getFill("unit1", callback="notempty")
        assert (2, 2, "map") not in test
        test = self.test.getFill("unit2", callback="sideways")
        assert all([i[0] == 4 for i in test])
        assert len(self.test.getFill("empty", callback="empty")) == 0
        test = self.test.getFill((0, 0, "two"), callback="empty", distance=1)
        assert len(test) == 3
        assert len(self.test.getFill(False, "empty")) == 0


class TestDirectPathing:
    def setup(self):
        path = os.path.dirname(__file__)
        self.test = udebs.battleStart(path + "/test.xml")
        self.one = self.test.map["one"]
        self.two = self.test.map["two"]
        self.map = self.test.map['map']

    def test_getAdjacent(self):
        test = list(self.two.getAdjacent({(0, 0, "two")}, callback=self.test["empty"], state=self.test))
        assert len(test) == 4
        assert len(test[0]) == 1
        test2 = list(self.two.getAdjacent({(0, 0, "two")}, callback=self.test["empty"], state=self.test))
        assert len(test2) == 4
        assert test2[0] == {(0, 0, "two")}

    def test_testLoc(self):
        assert self.two.testLoc((0, 0)) is False



import copy
import udebs
import os
from pytest import raises


class TestInstanceGets:
    def setup(self):
        path = os.path.dirname(__file__)
        self.env = udebs.battleStart(path + "/test.xml", log=True)

    def test_eq(self):
        assert self.env == self.env
        assert self.env != {}
        cp = copy.copy(self.env)
        assert cp == self.env
        self.env.controlIncrement("unit1", "ACT", 1)
        assert cp != self.env

        cp = copy.copy(self.env)
        cp.time += 1
        assert cp != self.env

        cp = copy.copy(self.env)
        cp["nope"] = 2
        assert self.env != cp

    def test_get_entity(self):
        unit = self.env["unit1"]

        assert unit is self.env.getEntity("unit1")
        assert unit is self.env.getEntity((0, 0, "two"))
        assert unit in self.env.getEntity(["unit1", (0, 0, "two")])

        assert self.env["empty"] == self.env.getEntity(False)

        with raises(udebs.UndefinedSelectorError):
            self.env.getEntity("unit")

        with raises(udebs.UndefinedSelectorError):
            self.env.getEntity((0, 0, "unknown"))

        with raises(udebs.UndefinedSelectorError):
            self.env.getEntity((-1, 0, "two"))

    def test_getMap(self):
        map_ = self.env.getMap("two")
        assert self.env.getMap(map_) == map_
        assert self.env.getMap((0, 0, "two")) == map_
        assert self.env.getMap(self.env["unit1"]) == map_
        with raises(udebs.UndefinedSelectorError):
            self.env.getMap("unknown")

    def test_castSingle(self):
        self.env.castSingle("unit1 ACT += 1")
        # This also tests rlist and groups
        assert self.env.getStat("unit1", "ACT") == 16

    def test_udebs_parser(self):
        test = self.env._getEntityUdebsStr("(function((unit).NAME),function(unit.LOC))")
        assert len(test.require) == 2

    # Entity controllers
    def test_getStat(self):
        with raises(udebs.UndefinedSelectorError):
            self.env.getStat("unit1", "nope")

        assert self.env.getStat("empty", "envtime") == 4
        assert self.env.getStat("unit1", "DESC") == "description"
        assert self.env.getStat("unit1", "group") == ["group1"]

    def test_controlListAdd(self):
        self.env.castSingle("unit2 inventory GETS move1")
        assert len(self.env.getStat("unit2", "inventory")) == 1

    def test_controlListRemove(self):
        self.env.castSingle("unit1 inventory LOSES move1")
        assert len(self.env.getStat("unit1", "inventory")) == 1

    def test_controlListClear(self):
        self.env.castSingle("unit1 CLEAR inventory")
        assert len(self.env.getStat("unit1", "inventory")) == 0

    def test_controlShuffle(self):
        self.env.castSingle("unit1 SHUFFLE inventory")
        assert len(self.env.getStat("unit1", "inventory")) == 2

    def test_controlIncrement(self):
        self.env.castSingle("unit2 ACT -= 1")
        assert self.env.getStat("unit2", "ACT") == -1

    def test_controlString(self):
        self.env.castSingle("unit2 ACT REPLACE 10")
        assert self.env.getStat("unit2", "ACT") == 10

    def test_controlRecruit(self):
        script = "unit2 RECRUIT (1 1 two)"
        self.env.castSingle(script)
        assert self.env.getEntity((1, 1, "two")).name == "unit21"

        self.env.castSingle("unit2 RECRUIT false")
        assert self.env.getEntity("unit22").loc is None

    def test_controlTravel(self):
        self.env.castSingle("unit1 MOVE unit2")
        assert self.env.getEntity("unit1").loc == (1, 0, "two")
        assert self.env.getEntity("unit2").loc is None
        assert self.env["empty"].loc is None

        with raises(Exception):
            self.env.castSingle("unit1 MOVE (FILL (0 0 two) empty)")

    # Entity Locs
    def test_getLoc(self):
        assert self.env.castSingle("unit1.LOC") == (0, 0, "two")

    def test_getX(self):
        assert self.env.castSingle("unit1.XLOC") == 0

    def test_getY(self):
        assert self.env.castSingle("unit1.YLOC") == 0

    def test_getName(self):
        assert self.env.castSingle("unit1.NAME") == "unit1"

    # Get Functions
    def test_getListStat(self):
        assert self.env.castSingle("unit1 inventory LISTSTAT ACT") == 5
        assert self.env.castSingle("unit1 inventory LISTSTAT inventory") == []
        assert self.env.castSingle("unit1 inventory LISTSTAT DESC") == ""

    def test_getListGroup(self):
        assert self.env.castSingle("unit1 inventory LISTGROUP move2") == "move1"
        assert self.env.castSingle("unit1 inventory LISTGROUP move1") is False

    def test_getSearch(self):
        assert len(self.env.castSingle("SEARCH group1")) == 1

    def test_getFilter(self):
        assert len(self.env.getSearch("group1")) == 1

    def test_gameLoop(self):
        index = 0
        self.env.cont = True
        for i in self.env.gameLoop():
            if index == 1:
                i.next = self.env

            index += 1
            if index > 4:
                i.castSingle("EXIT")

        assert index == 5
        assert self.env.time == 7
        self.env.resetState()
        assert self.env.time == 0
        assert self.env.cont is True

    def test_revert(self):
        self.env.controlTime(4)
        assert self.env.time == 8
        a = self.env.getRevert(2)
        assert a.time == 6
        assert self.env.getRevert(100) is False

    def test_error(self):
        error = udebs.UndefinedSelectorError("one", "entity")
        print(error)

    # Other functions
    def test_mapIter(self):
        assert len(list(self.env.mapIter("two"))) == 6

    def test_castAction(self):
        assert self.env.castAction("move2", "move2") is False

    def test_testMove(self):
        assert self.env.testMove("empty", "empty", "tick")
        assert self.env.testMove("empty", "empty", "move2") is False



from udebs.entity import Entity
import udebs
import os


class TestEntityClass:
    def setup(self):
        path = os.path.dirname(__file__)
        self.env = udebs.battleStart(path + "/test.xml")

    def test_init(self):
        new = Entity(self.env)
        stats = self.env.strings.union(self.env.stats, self.env.lists)
        for stat in stats:
            assert hasattr(new, stat)

    def test_copy(self):
        unit = self.env["unit1"]
        assert unit == unit.copy()

    def test_len(self):
        assert len(self.env["unit1"]) == 1
        assert len(list(iter(self.env["unit1"]))) == 1

    def test_clone(self):
        unit = self.env["unit1"]
        unit1 = unit.clone()

        assert unit.increment == 1
        assert unit1.increment == 0
        assert unit.name == "unit1"
        assert unit1.name == "unit11"

    def test_eq(self):
        unit = self.env["unit1"]
        assert (unit == "test") is False
        print(unit, repr(unit))



import copy
import json
import itertools
import os
import operator
from .errors import *


# ---------------------------------------------------
#            Imports and Variables                  -
# ---------------------------------------------------

class Standard:
    """
    Basic functionality wrappers.

    Do not import any of these, included only as reference for udebs config file syntax.
    """

    @staticmethod
    def print(*args):
        """
        prints extra output to console.

        .. code-block:: xml

            <i>print arg1 arg2 ...</i>
        """
        print(*args)
        return True

    @staticmethod
    def logicif(cond, value, other):
        """
        returns value if condition else other.

        TODO: Other is executed even if value is true.

        .. code-block:: xml

            <i>if cond value other</i>
        """
        return value if cond else other

    @staticmethod
    def inside(before, after, amount=1):
        """
        Returns true if before in after amount times else false.

        .. code-block:: xml

            <i>value in obj</i>
        """
        if isinstance(after, str):
            return before in after

        if amount == 0:
            return True

        count = 0
        for item in after:
            if item == before:
                count += 1
                if count >= amount:
                    return True

        return False

    @staticmethod
    def notin(*args, **kwargs):
        """
        Returns false if value in obj else true.

        .. code-block:: xml

            <i>value in obj</i>
        """
        return not Standard.inside(*args, **kwargs)

    @staticmethod
    def equal(*args):
        """Checks for equality of args.

        .. code-block:: xml

            <i>== arg1 arg2 ...</i>
            <i>arg1 == arg2</i>
        """
        x = args[0]
        for y in args:
            if y != x:
                return False
        return True

    @staticmethod
    def notequal(*args):
        """Checks for inequality of args.

        .. code-block:: xml

            <i>!= arg1 arg2 ...</i>
            <i>arg1 != arg2</i>
        """
        x = args[0]
        for y in args[1:]:
            if x == y:
                return False
        return True

    @staticmethod
    def gt(before, after):
        """Checks if before is greater than after

        .. code-block:: xml

            <i>before &gt; after</i>
        """
        return before > after

    @staticmethod
    def lt(before, after):
        """Checks if before is less than after

        .. code-block:: xml

            <i>before &lt; after</i>
        """
        return before < after

    @staticmethod
    def gtequal(before, after):
        """Checks if before is greater than or equal to after

        .. code-block:: xml

            <i>before &gt;= after</i>
        """
        return before >= after

    @staticmethod
    def ltequal(before, after):
        """Checks if before is less than or equal to after

        .. code-block:: xml

            <i>before &lt;= after</i>
        """
        return before <= after

    @staticmethod
    def plus(*args):
        """Sums arguments

        .. code-block:: xml

            <i>arg1 + arg2</i>
            <i>+ arg1 arg2 ...</i>
        """
        return sum(args)

    @staticmethod
    def multiply(*args):
        """Multiplies arguments

        .. code-block:: xml

            <i>arg1 * arg2</i>
            <i>* arg1 arg2 ...</i>
        """
        i = 1
        for number in args:
            i *= number
        return i

    @staticmethod
    def logicor(*args, storage=None, field=None):
        """
        returns true if even one of args is true.

        Note: All arguments are processed unless extra arguments are quoted.

        .. code-block:: xml

            <i>arg1 or arg2</i>
            <i>or arg1 arg2 ...</i>
        """
        env = _getEnv(storage, {"self": field})
        for i in args:
            if isinstance(i, UdebsStr):
                i = field.getEntity(i).testRequire(env)
            if i:
                return True
        return False

    @staticmethod
    def mod(before, after):
        """Returns before mod after.

        .. code-block:: xml

            <i>before % after</i>
        """
        return before % after

    @staticmethod
    def setvar(storage, variable, value):
        """Stores value inside of variable.

        Note: always returns true so can be used in require block.

        .. code-block:: xml

            <i>variable = value</i>
            <i>variable -> value</i>
        """
        storage[variable] = value
        return True

    @staticmethod
    def getvar(storage, variable):
        """Retrieves a variable

        .. code-block:: xml

            <i>$ variable</i>
            <i>$variable</i>
        """
        return storage[variable]

    @staticmethod
    def div(before, after):
        """Returns before divided by after.

        .. code-block:: xml

            <i>before / after</i>
        """
        return before / after

    @staticmethod
    def logicnot(element):
        """Switches a boolean from true to false and vice versa

        .. code-block:: xml

            <i>! element</i>
            <i>!element</i>
        """
        return not element

    @staticmethod
    def minus(before, element):
        """Returns before - element. (before defaults to 0 if not given)

        .. code-block:: xml

            <i>before - element</i>
            <i>-element</i>
        """
        return before - element

    @staticmethod
    def sub(array, i):
        """Gets the ith element of array.

        .. code-block:: xml

            <i>array sub i</i>
        """
        return next(itertools.islice(array, int(i), None), 'empty')

    @staticmethod
    def length(list_):
        """Returns the length of an iterable.

        .. code-block:: xml

            <i>length list_</i>
        """
        return len(list(list_))

    @staticmethod
    def quote(string):
        """Treats input as string literal and does not process commands.

        .. code-block:: xml
@staticmethod
            <i>`(caster CAST target move)</i>
        """
        return UdebsStr(string)


class Variables:
    versions = [0, 1]
    modules = {
        -1: {},
    }
    env = {
        "__builtins__": {"abs": abs, "min": min, "max": max, "len": len},
        "standard": Standard,
        "operator": operator,
        "storage": {},
    }
    default = {
        "f": "",
        "args": [],
        "kwargs": {},
        "all": False,
        "default": {},
        "string": [],
    }

    @staticmethod
    def keywords(version=1):
        return dict(Variables.modules[version], **Variables.modules[-1])


def importFunction(f, args):
    """
    Allows a user to import a single function into udebs.

    **deprecated - please use udebs.utilities.register
    """

    module = {
        f.__name__: {
            "f": f.__name__
        }
    }

    module[f.__name__].update(args)
    importModule(module, {f.__name__: f})


def importModule(dicts=None, globs=None, version=-1):
    """
    Allows user to extend base variables available to the interpreter.
    Should be run before the instance object is created.

    **deprecated for users - please use udebs.utilities.register
    """
    if globs is None:
        globs = {}
    if dicts is None:
        dicts = {}
    if version not in Variables.modules:
        Variables.modules[version] = {}

    Variables.modules[version].update(dicts)
    Variables.env.update(globs)


def importSystemModule(name, globs=None):
    """Convenience script for import system keywords."""
    if globs is None:
        globs = {}
    path = os.path.dirname(__file__)
    for version in Variables.versions:
        filename = "{}/keywords/{}-{}.json".format(path, name, str(version))
        with open(filename) as fp:
            importModule(json.load(fp), globs, version)


def _getEnv(local, glob=None):
    """Retrieves a copy of the base variables."""
    value = copy.copy(Variables.env)
    if glob:
        value.update(glob)
    value["storage"] = local
    return value


# ---------------------------------------------------
#            Interpreter Functions                 -
# ---------------------------------------------------
def formatS(string, version):
    """Converts a string into its python representation."""
    string = str(string)
    if string == "self":
        return string
    elif string == "false":
        return "False"
    elif string == "true":
        return "True"
    elif string == "None":
        return string
    elif string.isdigit():
        return string
    # String quoted by user.
    elif string[0] == string[-1] and string[0] in {"'", '"'}:
        return string
    # String has already been handled by call
    elif string[-1] == ")":
        return string
    elif string in Variables.env:
        return string
    # In case prefix notation used in keyword defaults.
    elif string[0] in Variables.keywords(version):
        return interpret(string, version)
    else:
        return "'" + string + "'"


def call(args, version):
    """Converts callList into functionString."""
    # Find keyword
    keywords = [i for i in args if i in Variables.keywords(version)]

    # Too many keywords is a syntax error.
    if len(keywords) > 1:
        raise UdebsSyntaxError("CallList contains to many keywords '{}'".format(args))

    # No keywords creates a tuple object.
    elif len(keywords) == 0:
        return "(" + ",".join(formatS(i, version) for i in args) + ")"

    keyword = keywords[0]

    # Get and fix data for this keyword.
    data = copy.copy(Variables.default)
    data.update(Variables.keywords(version)[keyword])

    # Create dict of values
    current = args.index(keyword)
    nodes = copy.copy(data["default"])

    for index in range(len(args)):
        value = "$" if index >= current else "-$"
        value += str(abs(index - current))
        if args[index] != keyword:
            nodes[value] = args[index]

    # Force strings into quoted arguments.
    for string in data["string"]:
        nodes[string] = "'" + str(nodes[string]).replace("'", "\\'") + "'"

    # Claim keyword arguments.
    kwargs = {}
    for key, value in data["kwargs"].items():
        if value in nodes:
            new_value = nodes[value]
            del nodes[value]
        else:
            new_value = value
        kwargs[key] = formatS(new_value, version)

    arguments = []
    # Insert positional arguments
    for key in data["args"]:
        if key in nodes:
            arguments.append(formatS(nodes[key], version))
            del nodes[key]
        else:
            arguments.append(formatS(key, version))

    # Insert ... arguments.
    if data["all"]:
        for key in sorted(nodes.keys(), key=lambda x: int(x.replace("$", ""))):
            arguments.append(formatS(nodes[key], version))
            del nodes[key]

    if len(nodes) > 0:
        raise UdebsSyntaxError("Keyword contains unused arguments. '{}'".format(" ".join(args)))

    # Insert keyword arguments.
    for key in sorted(kwargs.keys()):
        arguments.append(str(key) + "=" + str(kwargs[key]))

    return data["f"] + "(" + ",".join(arguments) + ")"


def split_callstring(raw, version):
    """Converts callString into call_list."""
    open_bracket = {'(', '{', '['}
    close_bracket = {')', '}', ']'}
    call_list = []
    buf = ''
    in_brackets = 0
    in_quotes = False
    dot_legal = True

    for char in raw.strip():

        if char in {'"', "'"}:
            in_quotes = not in_quotes

        elif not in_quotes:
            if char in open_bracket:
                in_brackets += 1

            elif char in close_bracket:
                in_brackets -= 1

            elif not in_brackets:
                if dot_legal:
                    if char == ".":
                        call_list.append(buf)
                        buf = ''
                        continue

                    elif char.isspace():
                        dot_legal = False
                        if call_list:
                            call_list = [".".join(call_list) + "." + buf]
                            buf = ''

                if char.isspace():
                    if buf:
                        call_list.append(buf)
                        buf = ''
                    continue

        buf += char
    call_list.append(buf)

    if in_brackets:
        raise UdebsSyntaxError("Brackets are mismatched. '{}'".format(raw))

    if '' in call_list:
        raise UdebsSyntaxError("Empty element in call_list. '{}'".format(raw))

    # Length one special cases.
    if len(call_list) == 1:
        value = call_list[0]

        # Prefix calling.
        if value not in Variables.keywords(version):
            if value[0] in Variables.keywords(version):
                return [value[0], value[1:]]

    return call_list


def interpret(string, version=1, debug=False):
    """Recursive function that parses callString"""
    try:
        _list = split_callstring(string, version)
        if debug:
            print("Interpret:", string)
            print("Split:", _list)

        found = []
        for entry in _list:
            if entry[0] == "(" and entry[-1] == ")":
                found.append(interpret(entry[1:-1], version, debug))
            elif "." in entry:
                found.append(interpret(entry, version, debug))
            elif entry[0] in Variables.keywords(version) and entry not in Variables.keywords(version):
                found.append(interpret(entry, version, debug))
            else:
                found.append(entry)

        comp = call(found, version)
        if debug:
            print("call:", _list)
            print("computed:", comp)

        return UdebsStr(comp)

    except Exception:
        print(string)
        raise


# ---------------------------------------------------
#                Script Main Class                 -
# ---------------------------------------------------
# An easy way to distinguish between interpreted strings.
class UdebsStr(str):
    pass


class Script:
    def __init__(self, effect, version=1, debug=False):
        # Raw text given to script.
        self.raw = effect
        self.interpret = effect
        if not isinstance(effect, UdebsStr):
            self.interpret = interpret(effect, version, debug)

        self.code = compile(self.interpret, '<string>', "eval")

    def __repr__(self):
        return "<Script " + self.raw + ">"

    def __str__(self):
        return self.raw

    def __call__(self, env):
        return eval(self.code, env)

    def __eq__(self, other):
        if not isinstance(other, Script):
            return False

        return self.raw == other.raw


# ---------------------------------------------------
#                     Runtime                      -
# ---------------------------------------------------
importSystemModule("base")
importSystemModule("udebs")



class UdebsError(Exception):
    """Base class for all Udebs errors."""
    pass


class UdebsSyntaxError(UdebsError):
    """Is raised when an effect or require is malformed and fails to parse."""

    def __init__(self, string):
        self.message = string

    def __str__(self):
        return repr(self.message)


class UdebsExecutionError(UdebsError):
    """Is raised when an error occurs during execution of a udebs action."""

    def __init__(self, script):
        self.script = script

    def __str__(self):
        return "invalid '{}' {}".format(self.script.raw, self.script.interpret)


class UndefinedSelectorError(UdebsError):
    """Is raised when udebs encounters an invalid reference to a udebs object."""

    def __init__(self, target, _type):
        self.selector = target
        self.type = _type

    def __str__(self):
        return "{} is not a valid {} selector.".format(repr(self.selector), self.type)



import traceback
import itertools
import time
import inspect
import functools
from collections import OrderedDict

from .interpret import importModule


# ---------------------------------------------------
#                  Utilities                       -
# ---------------------------------------------------

def cache(f=None, maxsize=None, storage=None):
    """Function decorator, lru_cache handling Instance objects as str(Instance)."""
    if maxsize is None:
        maxsize = float("inf")

    if storage is None:
        storage = OrderedDict()

    def cache_inside(f2):
        @functools.wraps(f2)
        def cache_wrapper(self, *args, **kwargs):
            key = (self.hash(), *args)
            value = storage.get(key, None)
            if value is None:
                value = f2(self, *args, **kwargs)
                storage[key] = value
            else:
                storage.move_to_end(key)

            while storage.__len__() > maxsize:
                storage.popitem(False)

            return value

        return cache_wrapper

    if f is None:
        return cache_inside
    return cache_inside(f)


def modify_state(state, entities=None, logging=True, revert=True):
    """Experimental: Creates a copy of an Instance object with modifications designed for improved tree search."""
    if entities is None:
        entities = {}
    clone = state.copy()
    if logging:
        clone.logging = True

    if revert:
        clone.revert = 0
        clone.state = False

    if entities is not None:
        for key, attributes in entities.items():
            new = state[key].copy()

            for name, value in attributes.items():
                setattr(new, name, value)

            clone[key] = new

    return clone


@functools.total_ordering
class Result:
    """Experimental: Object for handling gt and lt relationships when counting turns to victory."""
    def __init__(self, value, turns):
        self.value = value
        self.turns = turns

    def __eq__(self, other):
        return (self.value, self.turns) == (other.value, other.turns)

    def __lt__(self, other):
        if self.value != other.value:
            return self.value < other.value
        elif self.value < 0:
            return self.turns < other.turns
        return self.turns > other.turns

    def __repr__(self):
        return str((self.value, self.turns))

    def __int__(self):
        return self.value


class Timer:
    """Basic Timing context manager. Prints out the time it takes it's context to close."""

    def __init__(self, verbose=True, name=""):
        self.verbose = verbose
        self.total = None
        self.name = name

    def __enter__(self, verbose=True):
        self.time = time.perf_counter()
        return self

    def __exit__(self, *args, **kwargs):
        self.total = time.perf_counter() - self.time
        if self.verbose:
            print("({})".format(self.name), "total time:", self.total)

    def __str__(self):
        return str(self.total)


def count_recursion(f):
    """Function decorator, prints how many times a function calls itself."""
    start_time = 0

    @functools.wraps(f)
    def count_recursion_wrapper(*args, **kwargs):
        nonlocal start_time
        p = (start_time == 0)
        start_time += 1
        r = f(*args, **kwargs)
        if p:
            print("nodes visited:", start_time)
            start_time = 0
        return r

    return count_recursion_wrapper


countrecursion = count_recursion


def no_recurse(f):
    """Wrapper function that forces a function to return True if it recurse."""

    def func(*args, **kwargs):
        for i in traceback.extract_stack():
            if i[2] == f.__name__:
                return True

        return f(*args, **kwargs)

    return func


norecurse = no_recurse


def alternate(*args):
    """An alternation function for processing moves."""
    processed = []
    for i in args:
        if not isinstance(i, list):
            i = [i]
        processed.append(i)

    maximum = max(len(i) for i in processed)
    gen = (itertools.islice(itertools.cycle(i), maximum) for i in processed)
    yield from zip(*gen)


def register(f, args=None, name=None):
    """Register a function with udebs. Works as a function or a decorator.

    .. code-block:: python

        @udebs.register({"args": ["$1", "$2", "$3"]})
        def TEST(arg1, arg2, arg3):
            return "hello world"

        def TEST2(arg1, arg2, arg3):
            return "hello world"

        udebs.register(TEST2, {"args": ["$1", "$2", "$3"]})

        @udebs.register({"args": ["$1", $2, $3]})
        class Test3:
            def __init__(self):
                self.message = "hello "

            def __call__(self, world):
                return self.message + world


    .. code-block:: xml

        <i>TEST one two three</i>

    """

    def wrapper(args2, f2):
        f_name = f2.__name__ if name is None else name

        if isinstance(args2, list):
            args2 = {"args": args2}

        importModule({
            f_name: {
                "f": f_name,
                **args2,
            }
        }, {f_name: f2() if inspect.isclass(f2) else f2})
        return f2

    if args is None:
        return functools.partial(wrapper, f)

    return wrapper(args, f)


# ---------------------------------------------------
#                  Deprecated                      -
# ---------------------------------------------------

def placeholder(name):
    """Register a placeholder function that will be allocated after the udebs instance is created.

    **deprecated - just register an empty function.
    """
    importModule({name: {
        "f": "f_" + name,
        "args": ["self"],
    }}, {"f_" + name: lambda x: None})


class Player:
    """Base Class for players.

    **deprecated - please use udebs.utilities.register
    """

    def __init__(self, name):
        importModule({name: {
            "f": "f_" + name,
            "args": ["self"],
        }}, {"f_" + name: self})

    def __call__(self, state):
        raise NotImplementedError("Player subclasses should implement a __call__ method.")


def lookup(name, table):
    """Function for adding a basic lookup function to the interpreter.

    ** deprecated - please use custom function.
    """

    def wrapper(*args):
        value = table
        for arg in args:
            try:
                value = value[arg]
            except KeyError:
                return 0

        return value

    importModule({name: {
        "f": "f_" + name,
        "all": True,
    }}, {"f_" + name: wrapper})
